//++++++++++++++++++++++++++++++++++
// +++++ DDERequest for MATLAB 16bit

#include <windows.h>
#include <ddeml.h>
#include <string.h>		//fuer memmove()
#include <stdlib.h>

#define DLLMEX
extern "C"
{
	#include "..\header\mex.h"
	#include "..\header\testargs.h"
}
#include "..\header\fastdde.h"

#pragma hdrstop


#define ARGS_IN_MIN 2
#define ARGS_IN_MAX 4

#define ARG1_IN		prhs[0]		//channel
#define ARG2_IN		prhs[1]		//item
#define ARG3_IN		prhs[2]		//FormatArray
#define ARG4_IN		prhs[3]		//timeout

#define ARG1_OUT	plhs[0]		//Data


extern "C" void mexFunction(
		  int     nlhs,
		  Matrix* plhs[],
		  int     nrhs,
		  Matrix* prhs[]
		  )
{
	DWORD	timeout			= TIMEOUT;
	WORD	cbFormat		= CF_TEXT;	//default
	BOOL	StrFormatChk    = TRUE;		//zweiten Wert im Format-Argument testen


	if ( nrhs >= ARGS_IN_MAX - 1 )							//3. Parameter (Clipboard-Format)?
		cbFormat = TestFormatArg( ARG3_IN, &StrFormatChk );
	if ( nrhs == ARGS_IN_MAX )								//4. Parameter (TimeOut)?		
		timeout = TestTimeoutArg( ARG4_IN );

	if (	TestIOArgs( nrhs, ARGS_IN_MIN, ARGS_IN_MAX ) &&		//alle erforderlichen Parameter angegeben?
			TestChannelArg( ARG1_IN ) &&						//Channel Argument ok?
			TestItemArg( ARG2_IN ) &&
			cbFormat != (WORD)-1 &&								//sind bei ungueltigen Argumenten
			timeout != (DWORD)-1	)							// -1
	{
		char*	szItem;
		DWORD	hChannel[2];	//2 mal 32 bit
		HLOCAL	hLocalItem;

		ReqFuncParam FParam;
		
        UINT n = (UINT)mxGetN(ARG2_IN) + 1;						//kein mxCalloc(), da spaeteres
		hLocalItem = LocalAlloc( LPTR, n );						//mxFree() Absturz verursacht
	    szItem = (char*) LocalLock( hLocalItem );
		mxGetString( ARG2_IN, szItem, n );
        
        hmemcpy( (char*)hChannel, (char huge*)mxGetPr(ARG1_IN), sizeof(double) );

		FParam.idInst		= hChannel[0];
		FParam.hConv		= hChannel[1];
		FParam.cbFormat		= cbFormat;
		FParam.timeout		= timeout;
		FParam.szItem		= szItem;

        FParam.DestMatrix	= &ARG1_OUT;

		
        if ( StrFormatChk )
        	CFTextStringReq( &FParam );		// String Matrix

        else
        {
	        if ( cbFormat == CF_TEXT )		// CF_TEXT Format
    	   		CFTextRequest( &FParam );
        	else                         	// Excel Format
				FastDDEReq( &FParam );
		}

		LocalUnlock( hLocalItem );
	}
}




/////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CFTextStringReq( ReqFuncParam* FParam )
{
    HDDEDATA	hData;
	HSZ			hszItem;

	hszItem = DdeCreateStringHandle( FParam->idInst, FParam->szItem, CP_WINANSI );
	if ( !hszItem )
		return TRUE;

	hData = DdeClientTransaction( (LPBYTE)NULL, (DWORD)0, FParam->hConv, hszItem,
															FParam->cbFormat, XTYP_REQUEST, FParam->timeout, NULL );
	if ( !hData )	
		return TRUE;

	char huge* hDataString = (char huge*)DdeAccessData( hData, NULL );

	*FParam->DestMatrix = mxCreateString( hDataString );

	return FALSE;
}



/////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CFTextRequest( ReqFuncParam* FParam )
{
    HDDEDATA	hData;
	HSZ			hszItem;
	HGLOBAL		hGlobal;
	int rows, columns;


	hszItem = DdeCreateStringHandle( FParam->idInst, FParam->szItem, CP_WINANSI );
	if ( !hszItem )
		return TRUE;

	hData = DdeClientTransaction( (LPBYTE)NULL, (DWORD)0, FParam->hConv, hszItem,
															FParam->cbFormat, XTYP_REQUEST, FParam->timeout, NULL );
	if ( !hData )	
		return TRUE;
	
	char huge* hDataString = (char huge*)DdeAccessData( hData, NULL );

	GetMatrixSize( hDataString, &rows, &columns );
	DWORD matrixSize = (DWORD)rows * (DWORD)columns;

//Speicher reservieren
	hGlobal = GlobalAlloc( GPTR, matrixSize * sizeof(double) );
	if ( !hGlobal )
	{
		DdeUnaccessData( hData );
		DdeFreeDataHandle( hData );
		DdeFreeStringHandle ( FParam->idInst, hszItem );
		return TRUE;
	}

    double huge* dest	= (double huge*) GlobalLock( hGlobal );	//Speicherbereich mit double-Werten
	if ( Str2Double( dest, hDataString ) )				//komplette Konvertierung string -> double
	{

//Matrix erzeugen
		*FParam->DestMatrix = mxCreateFull( rows, columns, REAL );

	    DWORD index;
		double value;
		double huge* src	= dest;		//die zu double konvertierten strings sind source
		dest				= (double huge*)mxGetPr( *FParam->DestMatrix ); 
	
//Matrix transponieren
		for ( int i = 0; i < columns; i++ )
		{
			for ( int k = 0; k < rows; k++ )
			{
	            index = (DWORD)i+ ((DWORD)k * (DWORD)columns);
				value = src[index];
				hmemcpy( (char huge*)dest, (char*)&value, sizeof(double) );
				dest++;
			}
		}
	}

	GlobalUnlock( hGlobal );
	GlobalFree( hGlobal );

	DdeUnaccessData( hData );
	DdeFreeDataHandle( hData );
	DdeFreeStringHandle ( FParam->idInst, hszItem );

	return FALSE;
}



/////////////////////////////////////////////////////////////////////////////////////////////////////
void GetMatrixSize( char huge* hDataString, int* rows, int* columns )
{
	DWORD i = 0;

	*rows = 0;		//Initialwerte
	*columns = 1;	//Initialwerte

	while ( hDataString[i++] )
	{
		if ( hDataString[i] == '\t' && *rows == 0 )		//Tabulatoren nur in der 1. Zeile der Matrix erfassen
			(*columns)++;								//und Anzahl der Spalten ermitteln
	    if ( hDataString[i] == '\r' )					//wenn Zeilenende, dann eine Zeile mehr
		    (*rows)++;
	}
}



/////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL Str2Double( double huge* dest, char huge* hDataString )
{
	DWORD i = 0;
	
	while ( *hDataString != '\0' )
	{
		if ( *hDataString == '\t' || *hDataString == '\r' )		//wenn im angegebenen Bereich (z.B. Z1S1:Z5S5)
		{														//ein Wert fehlt
			mexPrintf( "Error. Possible empty element of source matrix.\r\n" );
			return FALSE;
		}
		else	
			dest[i] = atof( hDataString );
		
		Go2Next( &hDataString );				//Zeiger auf naechste Zahl setzen
		i++;	
	}

    return TRUE;
}



/////////////////////////////////////////////////////////////////////////////////////////////////////
_inline void Go2Next( char huge* huge* hDataString )
{
	do
	{
		switch ( **hDataString )
		{
			case '\t':	(*hDataString)++;		//Tabulator ueberspringen -> Beginn naechste Zahl
						return;
			case '\r':	(*hDataString) += 2;	//CR+LF ueberspringen -> Beginn naechste Zahl
						return;
		}
	}
	while ( (*hDataString)++ );
}







///////////////////////////////////////////////////////DDE-Excel Format
BOOL FastDDEReq( ReqFuncParam* FParam )
{
    HDDEDATA	hData;
	HSZ			hszItem;
	
	TableHeader	TabHeader;
	

	hszItem = DdeCreateStringHandle( FParam->idInst, FParam->szItem, CP_WINANSI );
	if ( !hszItem )
		return TRUE;

	hData = DdeClientTransaction( (LPBYTE)NULL, (DWORD)0, FParam->hConv, hszItem,
															FParam->cbFormat, XTYP_REQUEST, FParam->timeout, NULL );
	BYTE* hDataBlock = DdeAccessData( hData, NULL );
	
//tdtTable-Header lesen
    hmemcpy( (char*)&TabHeader, (char*)hDataBlock, sizeof(TabHeader) );
    hDataBlock += sizeof(TabHeader);

//tdtFloat-Header entfernen
	if ( !RemoveHeader( &TabHeader, hDataBlock ) )
	{
		DdeUnaccessData( hData );
		DdeFreeStringHandle ( FParam->idInst, hszItem );
		DdeFreeDataHandle( hData );

    	return TRUE;
    }

//Matrix erzeugen
	*FParam->DestMatrix = mxCreateFull( (int)TabHeader.rows, (int)TabHeader.columns, REAL );


    DWORD index;
	double value;
	double huge* src	= (double huge*)hDataBlock;
	double huge* dest	= (double huge*)mxGetPr( *FParam->DestMatrix ); 
	
//	*** Matrix transponieren
	for ( int i = 0; i < (int)TabHeader.columns; i++ )
	{
		for ( int k = 0; k < (int)TabHeader.rows; k++ )
		{
            index = (DWORD)i+ ((DWORD)k * (DWORD)TabHeader.columns);
			value = src[index];
			hmemcpy( (char huge*)dest, (char*)&value, sizeof(double) );
			dest++;
		}
	}

	DdeUnaccessData( hData );

	DdeFreeStringHandle ( FParam->idInst, hszItem );
	DdeFreeDataHandle( hData );


	return FALSE;
}


BOOL RemoveHeader( TableHeader* TabHeader, BYTE huge* src )
{
//Groesse der Matrix in bytes
	DWORD		Size		= (DWORD)TabHeader->rows * (DWORD)TabHeader->columns * sizeof(double);

	DWORD		BytesMoved	= (DWORD)0;
	BOOL		result		= TRUE;
	BYTE huge*	dest		= src;	//der gleiche Speicherbereich wird benutzt

	FloatHeader	FtHeader;

	while ( BytesMoved < Size )
    {
	    hmemcpy( (char*)&FtHeader, (char*)src, sizeof(FtHeader) );	//FtHeader-Struktur laden
        src += sizeof(FtHeader);									//Anfang der double-Werte
		if ( FtHeader.Float != tdtFloat )							//wenn keine Float-Zahlen
		{
			result = FALSE;
			break;
		}

		memmove( (char*)dest, (char*)src, FtHeader.cbFloat);		//tdtFloat-Header mit nachfolgenden double-Werten
                                                                	//ueberschreiben
        src += FtHeader.cbFloat;									//Zeiger aktualisieren
        dest += FtHeader.cbFloat;
        BytesMoved += FtHeader.cbFloat;
    }
    
    return result;
}
